#include "CAalgos.h"

#include <algorithm>
#include <cassert>

typedef pair<list<TGraphNeighbor>::iterator, list<TGraphNeighbor>::iterator> TUndLink;

bool operator<(const TUndLink& x, const TUndLink& y) { return true; }

struct PEcard
{
	bool operator()(const set<TUndLink> &x, const set<TUndLink> &y)
	{
		return x.size() > y.size();
	}
};



void IATC::replace(Topology *T, TGraphNeighborIterator u, TGraphNeighborIterator v, unsigned int k, unsigned int k1)
{
	/* replace k1 with k */
// 	RadioIface *i;
// 	for (i=u->info->radio; i->channel!=k1; i++);
// 	i->channel = k;
	for (int i = 0; i < u->Info()->GetNumRadios(); i++)
	  if (u->Info()->Radio(i)->Channel() == k1)
	  {
	    SetChannel (u->Info()->Radio(i), k);
	    break;
	  }

	/* assign u->radio and v->radio */
// 	for (u->radio=&v->info->radio[0]; u->radio->channel != k; u->radio++);
	SetRadio (u, &(*v->Info()->Radio (v->Info()->findRadioWithChannel(k))));
// 	for (v->radio=&u->info->radio[0]; v->radio->channel != k; v->radio++);
	SetRadio (v, &(*u->Info()->Radio (u->Info()->findRadioWithChannel(k))));

	for (TGraphNeighborIterator w=T->Node(u->Id())->NeighborBegin(); w!=T->Node(u->Id())->NeighborEnd(); w++)
		if ((w->Radio()!=NULL) && (w!=v) && (w->RadioChannel() == k))		/* link already considered */
		{
			set<unsigned int> uChannels, wChannels, CommonChannels;

			for (unsigned int r=0; r<u->Info()->numAvailRadios(); uChannels.insert(u->Info()->Radio(r++)->Channel()));
			for (unsigned int r=0; r<w->Info()->numAvailRadios(); wChannels.insert(w->Info()->Radio(r++)->Channel()));

			/* compute the intersection between uChannels and wChannels */
			set_intersection(uChannels.begin(), uChannels.end(), wChannels.begin(), wChannels.end(), inserter(CommonChannels, CommonChannels.begin()));
			CommonChannels.erase(0);

			TGraphNeighborIterator z;
			for (z=T->Node(w->Id())->NeighborBegin(); z->Id() != u->Id(); z++);

			if (!CommonChannels.empty())
			{
// 				for (w->radio=&u->info->radio[0]; w->radio->channel != *(CommonChannels.begin()); w->radio++);
				SetRadio (w, &(*u->Info()->Radio (u->Info()->findRadioWithChannel(*(CommonChannels.begin())))));
// 				for (z->radio=&w->info->radio[0]; z->radio->channel != *(CommonChannels.begin()); z->radio++);
				SetRadio (z, &(*w->Info()->Radio (w->Info()->findRadioWithChannel(*(CommonChannels.begin())))));
			}
			else
				replace(T,w,z,k,k1);
		}
}


/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
double IATC::Run (Topology *T)
{
//	Topology* T = new Topology(*Gi);     /* copy of the initial graph */

	/* for each link find the set of potentially interfering links */
	multimap<set<TUndLink>, TUndLink, PEcard> PE;		/* the map is sorted in descending order of |PE[e]| */

	for (int u = 0; u < T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
			if (u < v->Id())		/* don't count a link twice */
			{
				TGraphNeighborIterator node2;
				for (node2=T->Node(v->Id())->NeighborBegin(); node2!=T->Node(v->Id())->NeighborEnd() && node2->Id()!=u; node2++);
				assert( node2 != T->Node(v->Id())->NeighborEnd() );

				TUndLink e(v,node2);
				set<TUndLink> PEe;
/* if i<->j potentially interfere with e, PEe only contains the pair <i,j> such that i->info->id > j->info->id */

				/* add all the links outgoing the nodes in the interference range of any of the end nodes of e */
				for (int x=0; x < T->GetNumNodes(); x++)
					for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
						if ( T->LinkItrs(T->Node(u),v).PotInterferesWith(T->LinkItrs(T->Node(x),y)) )
						{
							for (node2=T->Node(y->Id())->NeighborBegin(); node2!=T->Node(y->Id())->NeighborEnd() && node2->Id()!=x; node2++);
							assert( node2!=T->Node(y->Id())->NeighborEnd() );

							if ( y->Id() > node2->Id() )
								PEe.insert(TUndLink(y,node2));
							else
								PEe.insert(TUndLink(node2,y));
						}

				PE.insert(pair<set<TUndLink>, TUndLink>(PEe, e));
			}

	for (multimap<set<TUndLink>, TUndLink, PEcard>::const_iterator e=PE.begin(); e!=PE.end(); ++e)
	{
		set<unsigned int> uChannels, vChannels, CommonChannels;
		TGraphNeighborIterator u=e->second.first, v=e->second.second;	/* end nodes of e */

		for (unsigned int r=0; r<u->Info()->numAvailRadios(); uChannels.insert(u->Info()->Radio(r++)->Channel()));
		uChannels.erase(0);
		for (unsigned int r=0; r<v->Info()->numAvailRadios(); vChannels.insert(v->Info()->Radio(r++)->Channel()));
		vChannels.erase(0);
		CommonChannels.clear();

		/* compute the intersection between uChannels and vChannels */
		set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(CommonChannels, CommonChannels.begin()));

		if (!CommonChannels.empty())
		{	/* set u->radio and v->radio */
// 			for (u->radio=&v->info->radio[0]; u->radio->channel != *(CommonChannels.begin()); u->radio++);
			SetRadio (u, &(*v->Info()->Radio (v->Info()->findRadioWithChannel(*(CommonChannels.begin())))));
// 			for (v->radio=&u->info->radio[0]; v->radio->channel != *(CommonChannels.begin()); v->radio++);
			SetRadio (v, &(*u->Info()->Radio (u->Info()->findRadioWithChannel(*(CommonChannels.begin())))));
		}
		else if ( (CommonChannels.empty()) && (uChannels.size()<u->Info()->numAvailRadios()) && (vChannels.size()<v->Info()->numAvailRadios()) )
		{
			vector<int> occurrences(WMS.numAvailChannels()+1);
			for (vector<int>::iterator i=occurrences.begin(); i!=occurrences.end(); i++) *i=0;

			/* find the number of occurrences of each channel in PE[e] */
			for (set<TUndLink>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if (l->first->Radio() != NULL)
					occurrences[l->first->RadioChannel()]++;

			/* find the least used channel in PE[e] */
			unsigned int k = min_element(++occurrences.begin(), occurrences.end()) - occurrences.begin();
			assert(k>0);

			/* assign k to the first unassigned interface on both end nodes */
			if (uChannels.find(k) == uChannels.end())
				SetChannel (u->Info()->Radio(uChannels.size()), k);
			if (vChannels.find(k) == vChannels.end())
				SetChannel (v->Info()->Radio(vChannels.size()), k);

			/* set u->radio and v->radio */
// 			for (u->radio=&v->info->radio[0]; u->radio->channel != k; u->radio++);
			SetRadio (u, &(*v->Info()->Radio (v->Info()->findRadioWithChannel(k))));
// 			for (v->radio=&u->info->radio[0]; v->radio->channel != k; v->radio++);
			SetRadio (v, &(*u->Info()->Radio (u->Info()->findRadioWithChannel(k))));
		}
		else if ((CommonChannels.empty()) && (uChannels.size()<u->Info()->numAvailRadios()) && (vChannels.size()==v->Info()->numAvailRadios()) )
		{
			vector<int> occurrences(vChannels.size());
			for (vector<int>::iterator i=occurrences.begin(); i!=occurrences.end(); i++) *i=0;

			/* find the number of occurrences of the channels of v in PE[e] */
			set<unsigned int>::iterator pos;
			int j;
			for (set<TUndLink>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if ( (l->first->Radio() != NULL) && ((pos=vChannels.find(l->first->RadioChannel())) != vChannels.end()) )
				{
					j=0;
					for (set<unsigned int>::iterator temp=vChannels.begin(); temp!=pos; temp++)
						j++;
					occurrences[j]++;
				}

			/* find the least used channel in PE[e] */
			j = min_element(occurrences.begin(), occurrences.end()) - occurrences.begin();
			pos = vChannels.begin();
			for (int i=0; i<j; i++) pos++;
			unsigned int k = *pos;
			assert(k>0);

			/* assign k to the first unassigned interface on u->info->id */
			SetChannel (u->Info()->Radio(uChannels.size()), k);

			/* set u->radio and v->radio */
// 			for (u->radio=&v->info->radio[0]; u->radio->channel != k; u->radio++);
			SetRadio (u, &(*v->Info()->Radio (v->Info()->findRadioWithChannel(k))));
			SetRadio (v, &(*u->Info()->Radio(uChannels.size())));
		}
		else if ((CommonChannels.empty()) && (uChannels.size()==u->Info()->numAvailRadios()) && (vChannels.size()<v->Info()->numAvailRadios()) )
		{
			vector<int> occurrences(uChannels.size());
			for (vector<int>::iterator i=occurrences.begin(); i!=occurrences.end(); i++) *i=0;

			/* find the number of occurrences of the channels of u in PE[e] */
			set<unsigned int>::iterator pos;
			int j;
			for (set<TUndLink>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if ( (l->first->Radio() != NULL) && ((pos=uChannels.find(l->first->RadioChannel())) != uChannels.end()) )
				{
					j=0;
					for (set<unsigned int>::iterator temp=uChannels.begin(); temp!=pos; temp++)
						j++;
					occurrences[j]++;
				}

			/* find the least used channel in PE[e] */
			j = min_element(occurrences.begin(), occurrences.end()) - occurrences.begin();
			pos = uChannels.begin();
			for (int i=0; i<j; i++) pos++;
			unsigned int k = *pos;
			assert(k>0);

			/* assign k to the first unassigned interface on v->info->id */
			SetChannel (v->Info()->Radio(vChannels.size()), k);

			/* set u->radio and v->radio */
// 			for (v->radio=&u->info->radio[0]; v->radio->channel != k; v->radio++);
			SetRadio (v, &(*u->Info()->Radio (u->Info()->findRadioWithChannel(k))));
			SetRadio (u, &(*v->Info()->Radio(vChannels.size())));
		}
		else
		{
			vector<int> occurrences(uChannels.size()+vChannels.size());
			for (vector<int>::iterator i=occurrences.begin(); i!=occurrences.end(); i++) *i=0;

			/* find the number of occurrences of the channels of u and v in PE[e] */
			set<unsigned int>::iterator pos;
			int j;
			for (set<TUndLink>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if (l->first->Radio() != NULL) 
				{
 					if ((pos=uChannels.find(l->first->RadioChannel())) != uChannels.end())
					{
						j=0;
						for (set<unsigned int>::iterator temp=uChannels.begin(); temp!=pos; temp++)
							j++;
						occurrences[j]++;
					}
					else if ((pos=vChannels.find(l->first->RadioChannel())) != vChannels.end())
					{
						j=uChannels.size();
						for (set<unsigned int>::iterator temp=vChannels.begin(); temp!=pos; temp++)
							j++;
						occurrences[j]++;
					}
				}

			/* find the least used channel in PE[e] */
			unsigned int k,k1;
			j = min_element(occurrences.begin(), occurrences.end()) - occurrences.begin();
			if ((unsigned int)j<uChannels.size())
			{
				pos = uChannels.begin();
				for (int i=0; i<j; i++) pos++;
				k = *pos;
				/* find the most used channel in vChannels */
				vector<int>::iterator temp=occurrences.begin();
				for (unsigned int i=0; i<uChannels.size(); i++) temp++;
				j = max_element(temp, occurrences.end()) - temp;
				pos = vChannels.begin();
				for (int i=0; i<j; i++) pos++;
				k1 = *pos;
				replace(T,v,u,k,k1);
			}
			else
			{
				pos = vChannels.begin();
				for (int i=uChannels.size(); i<j; i++) pos++;
				k = *pos;
				/* find the most used channel in uChannels */
				vector<int>::iterator temp=occurrences.begin();
				for (unsigned int i=0; i<uChannels.size(); i++) temp++;
				j = max_element(occurrences.begin(), temp) - occurrences.begin();
				pos = uChannels.begin();
				for (int i=0; i<j; i++) pos++;
				k1 = *pos;
				replace(T,u,v,k,k1);
			}
		}
	}

	return 0;
}
