#include "WMtypes.h"

#include <set>
#include <map>
#include <algorithm>
#include <numeric>
#include <ext/functional>
#include <cassert>

/*
 * 
 *  DEPRECATED because uses undirected links
 * 
 */


/* computes the maxflow on the initial graph (given by T)
 * and records the flow on each link
 */
//void maxflow(Topology *T);        /* prototype */


/* Binary predicate needed to compare two pairs based on the second element (increasing order) */
struct less_data
{
	template <class Tpair>
	bool operator()(Tpair x, Tpair y)
	{
		return x.second < y.second;
	}
};


/* Binary predicate needed to compare two pairs based on the second element (decreasing order) */
struct more_data
{
	template <class Tpair>
	bool operator()(Tpair x, Tpair y)
	{
		return x.second > y.second;
	}
};


/* Binary predicate needed to compare two undirected links based on the second element */
inline bool less_flow(const TGraphNeighbor& x, const TGraphNeighbor& y) { return x.uLink->flow > y.uLink->flow; }

inline bool emptySet(const set<undirectedLink*>& x) { return x.empty(); }

inline double sumFlow(double result, undirectedLink *i) { return result + i->flow; }

inline bool minFlowSgc(const set<undirectedLink*>& x, const set<undirectedLink*>& y) { return accumulate(x.begin(), x.end(), 0., sumFlow) < accumulate(y.begin(), y.end(), 0., sumFlow); }



/* Performs the first stage of MCCA, the link-group binding 
   Returns the number of groups */
set<int> *Link_Group_Bind(Topology *T)
{
	int g = 1;       /* link group identifier */

	vector<TGraphNeighbor>::iterator h, t;

	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
	{
		/* find the set of groups assigned to u's links */
		set<int> G;
		for (list<TGraphNeighbor>::iterator w=u->neighbor.begin(); w!=u->neighbor.end(); w++)
			G.insert(w->uLink->group);
		G.erase(0);

		/* feasibility constraint violated */
		if (G.size() > u->info->numAvailRadios)
		{
			int min2;
			map<int, double>::iterator imin, inext, minl, minr;

			map<int,double> Ftot;
			for (set<int>::const_iterator i=G.begin(); i!=G.end(); i++)
				Ftot[*i]=0.;			/* initialization: do we need this? */

			/* compute Ftot for each group belonging to G */
			for (vector<TGraphNode>::const_iterator i=T->node.begin(); i!=T->node.end(); i++)
				for (list<TGraphNeighbor>::const_iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
					if ( G.find(j->uLink->group) != G.end() )
						Ftot[j->uLink->group] += j->uLink->flow / 2;	/* divided by 2 because each undirected
												   link is visited twice */
			/* perform mergings */
			for (int k=G.size()-u->info->numAvailRadios; k>0; k--)
			{
				/* determine the two groups with smallest Ftot */
				inext = min_element(Ftot.begin(), Ftot.end(), less_data());
				imin = inext++;
				if (inext != Ftot.end())
				{
					minr = min_element(inext, Ftot.end(), less_data());
					min2 = minr->first;
				}
				if (imin != Ftot.begin())
				{
					minl = min_element(Ftot.begin(), imin, less_data());
					min2 = minl->first;
				}
				if ( (inext != Ftot.end()) && (imin != Ftot.begin()) )
					min2 = min(*minl, *minr, less_data()).first;
				/* merge such two groups */
				for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
					for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
						if (j->uLink->group == imin->first)
						{
							j->uLink->group = min2;
							Ftot[min2] += j->uLink->flow;
						}
				/* remove group imin->first */
				G.erase(imin->first);
				Ftot.erase(imin);
			}
		}

		/* sort neighbors in descending order of flow */
		vector<TGraphNeighbor> temp;
		copy(u->neighbor.begin(), u->neighbor.end(), temp.begin());
		sort(temp.begin(), temp.end(), less_flow);
		h = temp.begin();
		t = --temp.end();

		map<int,double> Floc;
		for (set<int>::const_iterator i=G.begin(); i!=G.end(); i++)
			Floc[*i]=0.;			/* initialization: do we need this? */
		/* compute local flow for each group assigned to u's links */
		for (list<TGraphNeighbor>::iterator i=u->neighbor.begin(); i!=u->neighbor.end(); i++)
			if (i->uLink->group != 0)
				Floc[i->uLink->group] += i->uLink->flow;
		/* compute the total flow on all the links of u */
		double uFlow = 0.;
		for (list<TGraphNeighbor>::iterator i=u->neighbor.begin(); i!=u->neighbor.end(); i++)
			uFlow += i->uLink->flow;

		while (h <= t)
		{
			if (h->uLink->group == 0)
			/* assign a group to the element at the head */
			{
				if (G.size() < u->info->numAvailRadios)
				{
					h->uLink->group = g;
					G.insert(g);
					g++;
				}
				else
					h->uLink->group = min_element(Floc.begin(), Floc.end(), less_data())->first;
				Floc[h->uLink->group] += h->uLink->flow;
			}

			/* assign a group to the elements at the tail */
			while ( (Floc[h->uLink->group] < uFlow/u->info->numAvailRadios) && (h<t) )
			{
				if (t->uLink->group == 0)
				{
					t->uLink->group = h->uLink->group;
					Floc[h->uLink->group] += t->uLink->flow;
				}
				t--;
			}
			h++;
		}
	}

	/* due to group mergings, there are less than g-1 groups */
	set<int>* groups = new set<int>;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			groups->insert(v->uLink->group);

	return groups;
}


/*****
	 NOTA: il calcolo di groups si pu�spostare dalla fine di Link_Group_Bind all'inizio di Group_Channel_Assign,
	fondendolo col calcolo del largestFlow
 *****/


/* Performs the second stage of MCCA, the group-channel assignment */
map<int,unsigned int> *Group_Channel_Assign(Topology *T, set<int> *groups)
{
	unsigned int argmaxEcCard, maxEcCard;
	map<int, unsigned int>* Channel = new map<int, unsigned int>;		/* channel assigned to each group */

	vector< set<undirectedLink*> > E(T->numAvailChannels+1);
	for (vector< set<undirectedLink*> >::iterator i=E.begin(); i!=E.end(); i++)
		i->clear();

	/* compute for each group the largest flow on a link */
	vector< pair<int,double> > largestFlow;
	for (set<int>::iterator g=groups->begin(); g!=groups->end(); g++)
		largestFlow.push_back(pair<int,double>(*g,0.));

	vector<pair<int,double> >::iterator pos;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
		{
			pos = find_if(largestFlow.begin(), largestFlow.end(), __gnu_cxx::compose1(bind2nd(equal_to<int>(),v->uLink->group), __gnu_cxx::select1st<pair<int,double> >()) );

			assert( pos != largestFlow.end() );

			if ( v->uLink->flow > pos->second )
				pos->second = v->uLink->flow;
		}

	sort(largestFlow.begin(), largestFlow.end(), more_data());

	/* process groups in decreasing order of largest flow */
	for (vector< pair<int,double> >::iterator g=largestFlow.begin(); g!=largestFlow.end(); g++)
	{
		set<undirectedLink*> P;

		/* compute P(g) */
		for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
			for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				if ( v->uLink->group == g->first )   // consider links belonging to g
					for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
						for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
							if ( (T->PotInterfLinks(TLinkItrs(x,y),TLinkItrs(u,v))) && (y->uLink->group != g->first) )
								P.insert(y->uLink);
		/* compute the intersection of P(g) and Ec for all c */
		vector< set<undirectedLink*> > S(T->numAvailChannels+1);
		for (unsigned int i=1; i<=T->numAvailChannels; i++)
			set_intersection(P.begin(), P.end(), E[i].begin(), E[i].end(), inserter(S[i],S[i].begin()));

		vector< set<undirectedLink*> >::iterator emptySgc = find_if(++S.begin(), S.end(), emptySet);
		if ( emptySgc != S.end() )
		/* if there exists at least one c such that S(g,c) is empty */
		{
			/* find c such that S(g,c) is empty and |Ec| is maximum */
			argmaxEcCard = emptySgc-S.begin();
			maxEcCard = E[argmaxEcCard].size();
			while ( (emptySgc = find_if(++emptySgc, S.end(), emptySet)) != S.end() )
				if ( E[emptySgc-S.begin()].size() > maxEcCard )
				{
					argmaxEcCard = emptySgc-S.begin();
					maxEcCard = E[argmaxEcCard].size();
				}
			/* assign group g to such a c */
			(*Channel)[g->first] = argmaxEcCard;
		}
		else
		{
			(*Channel)[g->first] = min_element(++S.begin(), S.end(), minFlowSgc) - S.begin();
		}
		/* add all the links belonging to g to Ec */
		for (vector<TGraphNode>::const_iterator u=T->node.begin(); u!=T->node.end(); u++)
			for (list<TGraphNeighbor>::const_iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				if ( v->uLink->group == g->first)
					E[(*Channel)[g->first]].insert(v->uLink);
	}

	return Channel;			/* return the channel to group assignment */
}


/* Last step: assign channels to radios */
void Radio_Channel_Assign(Topology *T, map<int,unsigned int>* Channel)
{
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
	{
		set<int> groupSet;
		for (list<TGraphNeighbor>::const_iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			groupSet.insert(v->uLink->group);
		/* check that the number of different groups is less than the number of available radios */
		assert(groupSet.size() <= u->info->numAvailRadios);
		/* assign a channel to each radio */
		int i=0;
		for (set<int>::const_iterator g=groupSet.begin(); g!=groupSet.end(); g++)
			u->info->radio[i++].channel = (*Channel)[*g];

		/* make the radio field of each neighbor struct point to the info struct of the radio used to reach that neighbor */
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
		{
			RadioIface *r;
			for (r=u->info->radio; r->channel != (*Channel)[v->uLink->group]; r++);
			v->radio = r;
		}
	}
}



/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
void MCCA(Topology *T)
{
//	Topology* T = new Topology(*Gi);     /* copy of the initial graph */

	PFR_maxflow(T,true);

	Radio_Channel_Assign(T, Group_Channel_Assign(T, Link_Group_Bind(T)));

}
