#ifndef WIMESH_PARSERS
#define WIMESH_PARSERS

#include <libxml++/libxml++.h>
#include "WMtypes.h"



class WiMeshParser : public xmlpp::SaxParser
{
public:
  WiMeshParser ();
  virtual ~WiMeshParser();

  inline string getXmlTopoFile () { return xml_topo_file; }
  inline string getSerializedTopoFile () { return serialized_topo_file; }
  inline string getBuildGiMethod () { return buildGiMethod; }
  inline const list<pair<string,string> >& getBuildGiParams () { return buildGiParams; }
  inline string getPfrMethod () { return pfrMethod; }
  inline const list<pair<string,string> >& getPfrParams () { return pfrParams; }
  inline const map<int, NodePowerConsumptionProfile>& getNodePCprofile () { return nodePCprofile; }
  inline const map<int, RadioPowerConsumptionProfile>& getRadioPCprofile () { return radioPCprofile; }
  void setNodePCprofile (MeshNode* info);
  void setRadioPCprofile (RadioIface* radio);
  inline RndVar* getCDrv () { return CDrv; }
  inline double getCDseed () { return CDseed; }
  inline unsigned int getCDtype () { return CDtype; }
  inline string getCDid () { return CDid; }
  inline string getCAalgo() { return caAlgo; }
  inline const list<pair<string,string> >& getCAparams () { return caParams; }
  inline bool getpostCAopt () { return postCAopt; }


protected:
  string xml_topo_file;
  string serialized_topo_file;

  string buildGiMethod;
  list<pair<string,string> > buildGiParams;

  string pfrMethod;
  list<pair<string,string> > pfrParams;

  map<int, NodePowerConsumptionProfile> nodePCprofile;
  map<int, RadioPowerConsumptionProfile> radioPCprofile;

  RndVar* nodePCrv;
  double nodePCseed;
  RndVar* radioPCrv;
  double radioPCseed;
  MotherOfAll *nodeUrng, *radioUrng;

  RndVar* CDrv;
  double CDseed;
  unsigned int CDtype;
  string CDid;

  string caAlgo;
  list<pair<string,string> > caParams;

  bool postCAopt;

  //overrides:
  virtual void on_start_document();
  virtual void on_end_document();
  virtual void on_start_element(const Glib::ustring& name,
                                const AttributeList& properties);
  virtual void on_end_element(const Glib::ustring& name);
  virtual void on_characters(const Glib::ustring& characters);
  virtual void on_comment(const Glib::ustring& text);
  virtual void on_warning(const Glib::ustring& text);
  virtual void on_error(const Glib::ustring& text);
  virtual void on_fatal_error(const Glib::ustring& text);
  
};


#endif //WIMESH_PARSERS
