#ifndef WIMESH_RNDVARS
#define WIMESH_RNDVARS

#include <iostream>
#include <vector>
#include <numeric>
#include "newran.h"

class RndVar
{
private:
	vector<char> distros;
	vector<double> params;
	double *probs;
	string   name;

	void readDistro (istream& is);

public:
	Random *rv;

	/* constructor */
	RndVar(istream& is);
	
	/* destructor */
	~RndVar() { if (rv) delete rv; if (probs) free(probs); }

	string getName () { return name; }

	/* 
	 * This function provides random number generators with a
	 * random seed taken by current time
	 */
	static double getSeed ();

	void print(ostream& os);
};



/// Fake random number generator (takes values from a sequence)
class Sequence : public Random
{
   void operator=(const Sequence&) {}      // private so can't access
   vector<Real> values;
   unsigned int counter;

public:
   const char* Name();                    // identification
   Sequence(const vector<Real>& _values) : values(_values), counter(0) {}
   Real Next() { if (counter>=values.size()) counter=0; return values.at(counter++); }
   ExtReal Mean() const { return std::accumulate(values.begin(), values.end(), 0.)/values.size(); }
   ExtReal Variance() const { return 1.0/12.0; }       // fixme
   Real Density(Real x) const { return (x < 0 || x > 1) ? 0 : 1.0; }     // fixme
};


#endif

