#ifndef WIMESH_RTPROTO_NS3
#define WIMESH_RTPROTO_NS3

#include "wimesh-ns3.h"


/*
 * Given a WiMesh topology, configures an ns-3 simulation
 * using either Layer-2.5 (by default) or FLAME
 */

class WiMeshRtProtoNs3 : public WiMeshNs3
{
public:
  // valid stack values: "ns3::LayerStack", "ns3::FlameStack", "ns3::Dot11sStack" (default)
  WiMeshRtProtoNs3 (string _stack);
  
  WiMeshNs3* Clone () { return new WiMeshRtProtoNs3 (*this); }

#ifdef HAVE_LAYER2_5
  // enable the use of OLSR for Layer-2.5 (default = off)
  void EnableOlsr ();
#endif
  //virtual void PrintConfiguration ();

  /*
   * Takes care of switching channels to match the new topology.
   * Returns the number of radios changing channel
   */
  void ChangeTopology (Topology* T, int* changes = 0);
  
private:

//  void FindChannelIfIndexMap (int u, map<uint16_t, uint32_t>& m);
  
  /*
   * For each mesh node, create a wifi device for each channel it uses
   * Each wifi device has the IP address 10.1.<channel>.<meshnode_id+1>
   */
  virtual void CreateNs3MeshDevices ();
  
  virtual void ConfigureForwarding ();
  
  //virtual int DoFindNeighborByMacAddress (int me, Mac48Address neighbor, int& channel, Ptr<WifiNetDevice>* device = 0);

  /*
   * Notify L2.5 that node "me" is not able to reach node neighbor anymore (L2.5 at node "me" will not consider neighbor as potential 
   * next-hop).
   */
  virtual void DoReceiveFinalDataFailed (int me, Mac48Address neighbor);
  
  virtual void PrintConfiguration () {}
  
  //Set the kernel routes required to reach the hosts attached to the gateways
  void SetRoutes();

  // create a mesh interface without installing the mesh stack
  // useful, e.g., to create interfaces for radios turned off (channel 0), so that such
  // interfaces do not participate in the routing protocol
  // Inspired by ns3::MeshHelper::CreateInterface
  Ptr<WifiNetDevice> CreateInterfaceWithoutStack (const WifiPhyHelper &phyHelper, Ptr<Node> node, uint16_t channel);
  
#ifdef HAVE_LAYER2_5  
  void ConfigureL2_5Links ();
  
  void ConfigureL2_5Topology ();
  
  bool useOlsr;
#endif

  NetDeviceContainer meshPointDevices;
};


#endif
